package net.sf.saxon.composabletree;

import net.sf.saxon.event.Receiver;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.Type;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Value;

/**
  * CommentImpl is an implementation of a Comment node
  * @author Michael H. Kay
  */


final class CommentImpl extends NodeImpl {

  /**
   * An comment data.
   */
  protected static final class CommentData extends NodeData
  {
    /**
     * A comment content.
     */
    public String comment;

    /**
     * {@inheritDoc}
     */
    @Override
    public CommentImpl getNodeInfo(ParentNodeImpl parent, int index)
    {
      return new CommentImpl(parent, index, this);
    }
  }
  
    /**
     * Comment data.
     */
    protected final CommentData data;
    
    /**
     * Constructs a comment from a comment data.
     * @param parent a parent node.
     * @param index a node index.
     * @param data an instance of {@link CommentData}.
     */
    protected CommentImpl(ParentNodeImpl parent, int index, CommentData data) 
    {
      this.parent = parent;
      this.index = index;
      this.data = data;
    }

    public CommentImpl(String content) 
    {
      this.data = new CommentData();
      this.data.comment = content;
    }
    
    /**
     * Gets a comment data.
     */
    @Override 
    protected CommentData getNodeData() 
    {
      return data;
    }

    /**
     * {@inheritDoc}
     */
    @Override
    protected boolean canShareNodeData()
    {
      return true;
    }

    public final String getStringValue() {
        return data.comment;
    }

    /**
     * Get the typed value of this node.
     * Returns the string value, as an instance of xs:string
     */

    public SequenceIterator getTypedValue() {
        return SingletonIterator.makeIterator(new StringValue(getStringValue()));
    }

    /**
     * Get the typed value of this node.
     * Returns the string value, as an instance of xs:string
     */

    public Value atomize() {
        return new StringValue(getStringValue());
    }        

    public final int getNodeKind() {
        return Type.COMMENT;
    }

    /**
    * Copy this node to a given outputter
    */

    public void copy(Receiver out, int whichNamespaces, boolean copyAnnotations, int locationId) throws XPathException {
        out.comment(data.comment, locationId, 0);
    }


    /**
     * Replace the string-value of this node
     *
     * @param stringValue the new string value
     */

    public void replaceStringValue(CharSequence stringValue) 
    {
      data.comment = stringValue.toString();
    }
}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
