package net.sf.saxon.composabletree;

import net.sf.saxon.om.AxisIterator;
import net.sf.saxon.om.AxisIteratorImpl;
import net.sf.saxon.om.EmptyIterator;
import net.sf.saxon.om.FastStringBuffer;
import net.sf.saxon.om.Item;
import net.sf.saxon.om.Navigator;
import net.sf.saxon.om.NodeInfo;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingleNodeIterator;

import net.sf.saxon.pattern.AnyNodeTest;
import net.sf.saxon.pattern.NodeTest;

import net.sf.saxon.type.Type;

/**
  * ParentNodeImpl is an implementation of a non-leaf node (specifically, an Element node
  * or a Document node)
  * @author Michael H. Kay
  */


abstract class ParentNodeImpl extends NodeImpl {
  
  protected static abstract class ParentNodeData extends NodeData
  {
    public Object children = null;     // null for no children
                                          // a NodeData for a single child
                                          // a NodeData[] for >1 child
  }
  
  /**
   * Gets parent node data.
   */
  @Override
  protected abstract ParentNodeData getNodeData();
  
    /**
    * Determine if the node has any children.
    */

    public final boolean hasChildNodes() {
      ParentNodeData data = getNodeData();
      
        return (data.children!=null);
    }

    /**
     * Determine how many children the node has
     * @return the number of children of this parent node
     */

    public int getNumberOfChildren() {
      ParentNodeData data = getNodeData();
      
        if (data.children == null) {
            return 0;
        } else if (data.children instanceof NodeData) {
            return 1;
        } else {
            return ((NodeData[])data.children).length;
        }
    }

    /**
     * Get an enumeration of the children of this node
     * @param test A NodeTest to be satisfied by the child nodes, or null
     * if all child node are to be returned
     * @return an iterator over the children of this node
    */

    protected final AxisIterator enumerateChildren(NodeTest test) 
    {
      ParentNodeData data = getNodeData();
      
      if (data.children==null)
      {
        return EmptyIterator.getInstance();
      }
      else if (data.children instanceof NodeData) 
      {
        NodeData childData = (NodeData)data.children;
        NodeImpl child = childData.getNodeInfo(this, 0);

        if ((test == null) || (test instanceof AnyNodeTest))
        {
          return SingleNodeIterator.makeIterator(child);
        }
        else 
        {
          return Navigator.filteredSingleton(child, test);
        }
      } 
      else 
      {
        if ((test == null) || (test instanceof AnyNodeTest)) 
        {
          class Iterator extends AxisIteratorImpl
          {
            public Iterator(
              ParentNodeImpl parent, 
              NodeData[] children)
            {
              this.parent = parent;
              this.children = children;
            }
            
            @Override
            public SequenceIterator getAnother()
            {
              return new Iterator(parent, children);
            }

            @Override
            public Item next()
            {
              if (position >= children.length)
              {
                position = -1;
                current = null;
                
                return null;
              }
              
              if (position < 0)
              {
                return null;
              }
              
              NodeImpl node = children[position].getNodeInfo(parent, position);
              
              ++position;
              current = node;
              
              return node;
            }
            
            private ParentNodeImpl parent; 
            private NodeData[] children; 
          }

          NodeData[] children = (NodeData[])data.children;
          
          return new Iterator(this, children);
        } 
        else 
        {
          return new ChildEnumeration(this, test);
        }
      }
    }

    /**
    * Get the first child node of the element
    * @return the first child node of the required type, or null if there are no children
    */

    public final NodeInfo getFirstChild() 
    {
      ParentNodeData data = getNodeData();
      
      if (data.children==null) 
      {
        return null;
      }
      
      NodeData nodeData;
      
      if (data.children instanceof NodeData) 
      {
        nodeData = (NodeData)data.children;
      }
      else
      {
        NodeData[] children = (NodeData[])data.children;
        
        nodeData = children[0];
      }
      
      return nodeData.getNodeInfo(this, 0);
    }

    /**
    * Get the last child node of the element
    * @return the last child of the element, or null if there are no children
    */

    public final NodeInfo getLastChild() 
    {
      ParentNodeData data = getNodeData();
      
      if (data.children==null) 
      {
        return null;
      }
      
      NodeData nodeData;
      int index;
      
      if (data.children instanceof NodeData) 
      {
        nodeData = (NodeData)data.children;
        index = 0;
      }
      else
      {
        NodeData[] children = (NodeData[])data.children;
        
        index = children.length - 1;
        nodeData = children[index]; 
      }
        
      return nodeData.getNodeInfo(this, index);
    }

    /**
     * Get the nth child node of the element (numbering from 0)
     * @param n identifies the required child
     * @return the last child of the element, or null if there is no n'th child
    */

    protected final NodeImpl getNthChild(int n) 
    {
      ParentNodeData data = getNodeData();
      
      if (data.children==null) 
      {
        return null;
      }
      
      NodeData nodeData;
      
      if (data.children instanceof NodeData) 
      {
        if (n != 0)
        {
          return null;
        }
        
        nodeData = (NodeData)data.children;
      }
      else
      {
        NodeData[] children = (NodeData[])data.children;

        if ((n < 0) || (n >= children.length))
        {
          return null;
        }
        
        nodeData = children[n];
      }
      
      if (nodeData == null)
      {
        return null;
      }
      
      return nodeData.getNodeInfo(this, n);
    }

    /**
     * Remove a given child
     * @param child the child to be removed
     */
    protected void removeChild(NodeImpl child)
    {
      ParentNodeData data = getNodeData();
      
      if (data.children == null) 
      {
        return;
      }
      
      NodeData childData = child.getNodeData();

      if (data.children == childData) 
      {
        data.children = null;

        return;
      }
      
      NodeData[] children = (NodeData[])data.children;
      
      for(int i=0; i < children.length; i++) 
      {
        if (children[i] == childData) 
        {
          if (children.length == 2) 
          {
            data.children = children[1 - i];
          } 
          else 
          {
            NodeData[] n2 = new NodeData[children.length - 1];

            if (i > 0) 
            {
              System.arraycopy(children, 0, n2, 0, i);
            }
            
            if (i < children.length - 1)
            {
              System.arraycopy(children, i+1, n2, i, children.length - i - 1);
            }
            
            data.children = cleanUpChildren(n2);
          }

          break;
        }
      }
    }

    /**
     * Tidy up the children of the node. Merge adjacent text nodes; remove zero-length text nodes;
     * reallocate index numbers to each of the children
     * @param children the existing children
     * @return the replacement array of children
     */

    private NodeData[] cleanUpChildren(NodeData[] children) {
        boolean prevText = false;
        int j = 0;
        NodeData[] c2 = new NodeData[children.length];
        for (int i=0; i<children.length; i++) {
            NodeData node = children[i];
            if (node instanceof TextImpl.TextData) {
                TextImpl.TextData textData = (TextImpl.TextData)node;
              
                if (prevText) {
                    TextImpl.TextData prev = (TextImpl.TextData)c2[j-1];
                    
                    TextImpl.TextData newData = new TextImpl.TextData();
                    
                    newData.content = prev.content + textData.content;
                    c2[j-1] = newData;
                } else if (textData.content.length() > 0) {
                    prevText = true;
                    c2[j++] = node;
                }
            } else {
                c2[j++] = node;
                prevText = false;
            }
        }
        if (j == c2.length) {
            return c2;
        } else {
            NodeData[] c3 = new NodeData[j];
            System.arraycopy(c2, 0, c3, 0, j);
            return c3;
        }
    }


    /**
    * Return the string-value of the node, that is, the concatenation
    * of the character content of all descendent elements and text nodes.
    * @return the accumulated character content of the element, including descendant elements.
    */

    public String getStringValue() {
        return getStringValueCS().toString();
    }


    public CharSequence getStringValueCS() {
        FastStringBuffer sb = null;

        NodeImpl next = (NodeImpl)getFirstChild();
        while (next!=null) {
            if (next instanceof TextImpl) {
                if (sb==null) {
                    sb = new FastStringBuffer(FastStringBuffer.SMALL);
                }
                sb.append(next.getStringValueCS());
            }
            next = next.getNextInDocument(this);
        }
        if (sb==null) return "";
        return sb.condense();
    }

    /**
     * Supply an array to be used for the array of children. For system use only.
     * @param array the array to be used
    */

    protected void useChildrenArray(NodeData[] array) 
    {
      ParentNodeData data = getNodeData();
      
      data.children = array;
    }

    /**
     * Add a child node to this node. For system use only. Note: normalizing adjacent text nodes
     * is the responsibility of the caller.
     * @param node the node to be added as a child of this node
     * @param index the position where the child is to be added
    */

    public synchronized void addChild(NodeImpl node, int index) 
    {
      ParentNodeData data = getNodeData();
      NodeData[] c;

      if (data.children == null) 
      {
        c = new NodeData[10];
      } 
      else if (data.children instanceof NodeData) 
      {
        c = new NodeData[10];
        c[0] = (NodeData)data.children;
      }
      else 
      {
        c = (NodeData[])data.children;
      }
      
      if (index >= c.length) 
      {
        NodeData[] kids = new NodeData[c.length * 2];

        System.arraycopy(c, 0, kids, 0, c.length);
        c = kids;
      }
      
      NodeData nodeData = node.getNodeData();
      
      c[index] = nodeData;
      data.children = c;
    }


    /**
     * Insert copies of a sequence of nodes as children of this node.
     * <p/>
     * <p>This method takes no action unless the target node is a document node or element node. It also
     * takes no action in respect of any supplied nodes that are not elements, text nodes, comments, or
     * processing instructions.</p>
     * <p/>
     * <p>The supplied nodes will be copied to form the new children. Adjacent text nodes will be merged, and
     * zero-length text nodes removed.</p>
     *
     * @param source  the nodes to be inserted
     * @param atStart true if the new nodes are to be inserted before existing children; false if they are
     * @param inherit true if the inserted nodes are to inherit the namespaces that are in-scope for their
     * new parent; false if such namespaces should be undeclared on the children
     */

    public void insertChildren(NodeInfo[] source, boolean atStart, boolean inherit) {
        if (atStart) {
            insertChildrenAt(source, 0, inherit);
        } else {
            insertChildrenAt(source, getNumberOfChildren(), inherit);
        }
    }

    /**
     * Insert children before or after a given existing child
     * @param source the children to be inserted
     * @param index the position before which they are to be inserted: 0 indicates insertion before the
     * first child, 1 insertion before the second child, and so on.
     * @param inherit true if the inserted nodes are to inherit the namespaces that are in-scope for their
     * new parent; false if such namespaces should be undeclared on the children
     */

    protected synchronized void insertChildrenAt(
      NodeInfo[] source, 
      int index, 
      boolean inherit) 
    {
      ParentNodeData data = getNodeData();
      
      if (source.length == 0) 
      {
        return;
      }
      
      for(int i=0; i < source.length; i++) 
      {
        NodeImpl child = (NodeImpl)source[i];

        child.parent = this;
        
        if (child instanceof ElementImpl) 
        {
          // If the child has no xmlns="xxx" declaration, 
          // then add an xmlns="" to prevent false inheritance
          // from the new parent
          ElementImpl element = (ElementImpl)child;
          
          element.fixupInsertedNamespaces(inherit);
        }
      }
      
      if (data.children == null) 
      {
        if (source.length == 1) 
        {
          NodeImpl node = (NodeImpl)source[0];

          data.children = node.getNodeData();
        }
        else 
        {
          NodeData[] n2 = new NodeData[source.length];
          
          for(int i = 0; i < source.length; ++i)
          {
            NodeImpl node = (NodeImpl)source[i];

            n2[i] = node.getNodeData();
          }
          
          data.children = n2;
        }
      }
      else if (data.children instanceof NodeData) 
      {
        NodeData childData = (NodeData)data.children; 
        int adjacent = index == 0 ? 0 : source.length - 1;
        
        if ((childData instanceof TextImpl.TextData) && 
          (source[adjacent] instanceof TextImpl)) 
        {
          TextImpl.TextData childTextData = (TextImpl.TextData)childData; 
          TextImpl sourceText = (TextImpl)source[adjacent];
          
          if (index == 0) 
          {
            sourceText.replaceStringValue(
              source[adjacent].getStringValue() + childTextData.content);
          } 
          else 
          {
            sourceText.replaceStringValue(
              childTextData.content + source[adjacent].getStringValue());
          }
          
          NodeData[] n2 = new NodeData[source.length];
          
          for(int i = 0; i < source.length; ++i)
          {
            NodeImpl node = (NodeImpl)source[i];

            n2[i] = node.getNodeData();
          }
          
          data.children = n2;
        }
        else 
        {
          NodeData[] n2 = new NodeData[source.length + 1];
        
          if (index == 0) 
          {
            for(int i = 0; i < source.length; ++i)
            {
              NodeImpl node = (NodeImpl)source[i];

              n2[i] = node.getNodeData();
            }

            n2[source.length] = childData;
          } 
          else 
          {
            n2[0] = childData;

            for(int i = 0; i < source.length; ++i)
            {
              NodeImpl node = (NodeImpl)source[i];

              n2[i + 1] = node.getNodeData();
            }
          }
          
          data.children = cleanUpChildren(n2);
        }
      }
      else 
      {
        NodeData[] n0 = (NodeData[])data.children;
        NodeData[] n2 = new NodeData[n0.length + source.length];
        
        System.arraycopy(n0, 0, n2, 0, index);
        
        for(int i = 0; i < source.length; ++i)
        {
          NodeImpl node = (NodeImpl)source[i];

          n2[index + i] = node.getNodeData();
        }
        
        System.arraycopy(
          n0, 
          index, 
          n2, 
          index + source.length, 
          n0.length - index);
        
        data.children = cleanUpChildren(n2);
      }
    }

    /**
     * Replace child at a given index by new children
     * @param source the children to be inserted
     * @param index the position at which they are to be inserted: 0 indicates replacement of the
     * first child, replacement of the second child, and so on. The effect is undefined if index
     * is out of range
     * @param inherit set to true if the new child elements are to inherit the in-scope namespaces
     * of their new parent
     * @throws IllegalArgumentException if any of the replacement nodes is not an element, text,
     * comment, or processing instruction node
     */

    protected synchronized void replaceChildrenAt(
      NodeInfo[] source, 
      int index, 
      boolean inherit) 
    {
      ParentNodeData data = getNodeData();
      
      if (data.children == null) 
      {
        return;
      }
      
      for(int i=0; i < source.length; ++i) 
      {
        NodeImpl child = (NodeImpl)source[i];
        int kind = child.getNodeKind();
        
        child.parent = this;

        switch(kind) 
        {
          case Type.ELEMENT:
          {
            // If the child has no xmlns="xxx" declaration, 
            // then add an xmlns="" to prevent false inheritance
            // from the new parent
            ElementImpl element = (ElementImpl)child;
            
            element.fixupInsertedNamespaces(inherit);
         
            break;
          }
          case Type.COMMENT:
          case Type.TEXT:
          case Type.PROCESSING_INSTRUCTION:
          {
            break;
          }
          default:
          {
            throw new IllegalArgumentException(
              "Replacement child node is not an element, text, comment, or PI");
          }
        }
      }
      
      if (data.children instanceof NodeData) 
      {
        if (source.length == 0) 
        {
          data.children = null;
        } 
        else if (source.length == 1) 
        {
          NodeImpl node = (NodeImpl)source[0];
          
          data.children = node.getNodeData();
        } 
        else 
        {
          NodeData[] n2 = new NodeData[source.length];
          
          for(int i = 0; i < source.length; ++i)
          {
            NodeImpl node = (NodeImpl)source[i];
            
            n2[i] = node.getNodeData();
          }
          
          data.children = cleanUpChildren(n2);
        }
      } 
      else 
      {
        NodeData[] n0 = (NodeData[])data.children;
        NodeData[] n2 = new NodeData[n0.length + source.length - 1];
        
        System.arraycopy(n0, 0, n2, 0, index);
        
        for(int i = 0; i < source.length; ++i)
        {
          NodeImpl node = (NodeImpl)source[i];
          
          n2[index + i] = node.getNodeData();
        }
        
        System.arraycopy(
          n0, 
          index + 1, 
          n2, 
          index+source.length, 
          n0.length - index - 1);
        
        data.children = cleanUpChildren(n2);
      }
    }


    /**
     * Compact the space used by this node
     * @param size the number of actual children
     */

    public synchronized void compact(int size) {
      ParentNodeData data = getNodeData();
      
      if (size==0) 
      {
        data.children = null;
      } 
      else if (size==1) 
      {
        if (data.children instanceof NodeData[]) 
        {
          NodeData[] children = (NodeData[])data.children;
          
          data.children = children[0];
        }
      } 
      else 
      {
        NodeData[] kids = new NodeData[size];

        System.arraycopy(data.children, 0, kids, 0, size);
        data.children = kids;
      }
    }
}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
