package net.sf.saxon.composabletree;

import net.sf.saxon.event.Receiver;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.Type;
import net.sf.saxon.om.SequenceIterator;
import net.sf.saxon.om.SingletonIterator;
import net.sf.saxon.value.StringValue;
import net.sf.saxon.value.Value;

/**
  * ProcInstImpl is an implementation of ProcInstInfo used by the Propagator to construct
  * its trees.
  * @author Michael H. Kay
  */


class ProcInstImpl extends NodeImpl {
  
  /**
   * A processing instruction data.
   */
  protected static class ProcInstData extends NodeData
  {
    public String content;
    public int nameCode;
    public String systemId;
    public int lineNumber = -1;

    @Override
    public ProcInstImpl getNodeInfo(ParentNodeImpl parent, int index)
    {
      return new ProcInstImpl(parent, index, this);
    }
  }
  
  /**
   * A processing instruction data.
   */
  protected ProcInstData data;

  /**
   * Creates {@link ProcInstImpl} using {@link ProcInstData}.
   * @param parent a parent node.
   * @param index a node index.
   * @param data a {@link ProcInstData} instance.
   */
  public ProcInstImpl(ParentNodeImpl parent, int index, ProcInstData data)
  {
    this.parent = parent;
    this.index = index;
    this.data = data;
  }
  
  public ProcInstImpl(int nameCode, String content) 
  {
    this.data = new ProcInstData();
    this.data.nameCode = nameCode;
    this.data.content = content;
  }
    

  /**
   * Gets a {@link ProcInstData} instance.
   */
  @Override
  protected ProcInstData getNodeData()
  {
    return data;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected boolean canShareNodeData()
  {
    return true;
  }

  /**
	* Get the nameCode of the node. This is used to locate the name in the NamePool
	*/

	public int getNameCode() {
		return data.nameCode;
	}

    public String getStringValue() {
        return data.content;
    }
    /**
     * Get the typed value of this node.
     * Returns the string value, as an instance of xs:string
     */

    public SequenceIterator getTypedValue() {
        return SingletonIterator.makeIterator(new StringValue(getStringValue()));
    }

    /**
     * Get the typed value of this node.
     * Returns the string value, as an instance of xs:string
     */

    public Value atomize() {
        return new StringValue(getStringValue());
    }
    
    public final int getNodeKind() {
        return Type.PROCESSING_INSTRUCTION;
    }

    /**
     * Set the system ID and line number
     * @param uri the system identifier
     * @param lineNumber the line number
    */

    public void setLocation(String uri, int lineNumber) {
        this.data.systemId = uri;
        this.data.lineNumber = lineNumber;
    }

    /**
     * Get the system ID for the entity containing this node.
     * @return the system identifier
    */

    public String getSystemId() {
        return data.systemId;
    }

    /**
    * Get the line number of the node within its source entity
    */

    public int getLineNumber() {
        return data.lineNumber;
    }

    /**
    * Copy this node to a given outputter
    */

    public void copy(Receiver out, int whichNamespaces, boolean copyAnnotations, int locationId) throws XPathException {
        out.processingInstruction(getLocalPart(), data.content, locationId, 0);
    }

    // DOM methods

    /**
     * The target of this processing instruction. XML defines this as being
     * the first token following the markup that begins the processing
     * instruction.
     * @return the processing instruction name
     */

    public String getTarget() {
        return getLocalPart();
    }

    /**
     *  The content of this processing instruction. This is from the first non
     * white space character after the target to the character immediately
     * preceding the <code>?&gt;</code> .
     * @return the string value of the processing instruction node
     */

    public String getData() {
        return data.content;
    }


    /**
     * Rename this node
     *
     * @param newNameCode the NamePool code of the new name
     */

    public void rename(int newNameCode)  
    {
      // NOTE: its responsibility of caller to ensure that 
      // the node is not shared.
      data.nameCode = newNameCode;
    }


    /**
     * Replace the string-value of this node
     *
     * @param stringValue the new string value
     */
  
     public void replaceStringValue(CharSequence stringValue) 
     {
       // NOTE: its responsibility of caller to ensure that 
       // the node is not shared.
       data.content = stringValue.toString();
     }
}


//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//
