package net.sf.saxon.composabletree;

import net.sf.saxon.event.Receiver;
import net.sf.saxon.trans.XPathException;
import net.sf.saxon.type.Type;

/**
  * A node in the XML parse tree representing character content<P>
  * @author Michael H. Kay
  */

final class TextImpl extends NodeImpl {

  /**
   * A text data.
   */
  protected static final class TextData extends NodeData
  {
    public String content;

    /**
     * {@inheritDoc}
     */
    @Override
    public TextImpl getNodeInfo(ParentNodeImpl parent, int index)
    {
      return new TextImpl(parent, index, this);
    }
  }

  /**
   * A text data.
   */
  protected final TextData data;

  /**
   * Constructs a text node from a text data.
   * @param parent a parent node.
   * @param index a node index.
   * @param data a text data.
   */
  public TextImpl(ParentNodeImpl parent, int index, TextData data) 
  {
    this.parent = parent;
    this.index = index;
    this.data = data;
  }
  
  public TextImpl(ParentNodeImpl parent, String content) 
  {
  	this.parent = parent;
  	this.data = new TextData();
  	this.data.content = content;
  }
  
  /**
   * Gets a text data.
   */
  @Override
  protected TextData getNodeData() 
  {
    return data;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  protected boolean canShareNodeData()
  {
    return true;
  }

    /**
    * Return the character value of the node.
    * @return the string value of the node
    */

    public String getStringValue() {
      return data.content;
    }

    /**
    * Return the type of node.
    * @return Type.TEXT
    */

    public final int getNodeKind() {
        return Type.TEXT;
    }

    /**
    * Copy this node to a given outputter
    */

    public void copy(Receiver out, int whichNamespaces, boolean copyAnnotations, int locationId) throws XPathException {
        out.characters(data.content, locationId, 0);
    }


    /**
     * Replace the string-value of this node
     *
     * @param stringValue the new string value
     */

    public void replaceStringValue(CharSequence stringValue)
    {
      if (stringValue.length() == 0) 
      {
        delete();
      } 
      else 
      {
        data.content = stringValue.toString();
      }
    }
    
}

//
// The contents of this file are subject to the Mozilla Public License Version 1.0 (the "License");
// you may not use this file except in compliance with the License. You may obtain a copy of the
// License at http://www.mozilla.org/MPL/
//
// Software distributed under the License is distributed on an "AS IS" basis,
// WITHOUT WARRANTY OF ANY KIND, either express or implied.
// See the License for the specific language governing rights and limitations under the License.
//
// The Original Code is: all this file.
//
// The Initial Developer of the Original Code is Michael H. Kay.
//
// Portions created by (your name) are Copyright (C) (your legal entity). All Rights Reserved.
//
// Contributor(s): none.
//